#include "DatabaseWorker.hpp"

#include <QList>
#include <QMap>
#include <QSqlDatabase>
#include <QSqlError>
#include <QSqlIndex>
#include <QSqlQuery>
#include <QSqlRecord>
#include <QString>
#include <QVariant>
#include <QVector>

MySqlWorker::DatabaseWorker::DatabaseWorker(const QString &mySqlServer,
											const QString &database,
											const QString &user,
											const QString &password)
{
	QSqlDatabase db = QSqlDatabase::addDatabase("QMYSQL", database);
	db.setHostName(mySqlServer);
	db.setDatabaseName(database);
	db.setUserName(user);
	db.setPassword(password);
	databaseName = database;
}

MySqlWorker::DatabaseWorker::~DatabaseWorker() { }

QList<QVector<QString> *>* MySqlWorker::DatabaseWorker::searchEntries(const QString &table,
																	  const QMap<QString, QString> *filter,
																	  const QVector<QString> *columns,
																	  const QString sortBy,
																	  const bool sortAscending)
{
	QString queryString = "SELECT ";

	// which columns should appear in the result set?
	if (columns != nullptr)
	{
		for (int i = 0; i < columns->count(); ++i)
		{
			queryString += columns->at(i);
			if (i < columns->count() - 1) { queryString += ", "; }
		}
	}
	else
	{
		queryString += "* ";
	}

	// in which table (or expression) we should search?
	queryString += " FROM " + table;

	// with which conditions we should search?
	if (filter != nullptr)
	{
		queryString += " WHERE ";
		for (auto tuple = filter->begin(); tuple != filter->end(); ++tuple)
		{
			queryString += tuple.key() + " LIKE '%" + tuple.value() + "%'";
			if (tuple != filter->end() - 1)
			{
				queryString += " AND ";
			}
		}
	}

	// should we order the result?
	if (sortBy != "")
	{
		queryString += " ORDER BY " + sortBy;
		queryString += sortAscending ? " ASC" : " DSC";	// ordering ascending or descening?
	}

	// execution
	QList<QVector<QString>*> *res = new QList<QVector<QString>*>();

	QSqlDatabase db = QSqlDatabase::database(databaseName);
	QSqlQuery query(db);
	if (db.open()&& query.exec(queryString))
	{
		while (query.next())
		{
			int numberOfColumns = query.record().count();
			QVector<QString> *results = new QVector<QString>();
			for (int i = 0; i < numberOfColumns; ++i) { results->append(query.value(i).toString());}
			res->append(results);
		}
	}
	else
	{
		setLastError(query.lastError().text());
		delete res;
		res = nullptr;
	}
	db.close();
	return res;
}

bool MySqlWorker::DatabaseWorker::deleteEntry(const QString &table, const QMap<QString, QString> &filter)
{
	QString queryString = "DELETE FROM "  + table + " WHERE ";
	for (auto tuple = filter.begin(); tuple != filter.end(); ++tuple)
	{
		queryString += tuple.key() + "=" + "'" + tuple.value() + "'";
		if (tuple != filter.end() - 1) { queryString += " AND "; }
	}
	return executeMySQL(queryString);
}

bool MySqlWorker::DatabaseWorker::insertNewEntry(const QString &table, const QMap<QString, QString> &valueTuples)
{
	int numberOfCols = QSqlDatabase::database(databaseName).record(table).count();

	if (valueTuples.count() != numberOfCols)	// not enough values provided
	{
		lastError = tr("Error: Wrong number of values.\nExpected: %1.\nReceived: %2.")
					.arg(numberOfCols)
					.arg(valueTuples.count());
		return false;
	}

	#pragma region build query string

	QString queryString = "INSERT INTO " + table + " ";

	//columns
	QList<QString> columnNames = valueTuples.keys();
	queryString += "(";
	for (int i = 0; i < columnNames.count(); ++i)
	{
		queryString += columnNames[i];
		if (i < columnNames.count() - 1) { queryString += ","; }
	}
	queryString += ")";

	//values
	QList<QString> values = valueTuples.values();
	queryString += " VALUES (";
	for (int i = 0; i < values.count(); ++i)
	{
		queryString += values[i] == "NULL" ? "NULL" : "'" + values[i] + "'";
		if (i < values.count() - 1) { queryString += ","; }
	}
	queryString += ")";

	#pragma endregion

	return executeMySQL(queryString);
}

bool MySqlWorker::DatabaseWorker::updateEntries(const QString &table,
												const QMap<QString, QString> &filter,
												const QMap<QString, QString> &newVaules)
{
	QString queryString = "UPDATE " + table;

#pragma region query string part 2 (set)

	queryString += " SET ";
	for (auto tuple = newVaules.begin(); tuple != newVaules.end(); ++tuple)
	{
		queryString += tuple.key() + "=";
		queryString += tuple.value() == "NULL" ? "NULL" : "'" + tuple.value() + "'";
		if (tuple != newVaules.end() - 1) { queryString += ","; }
	}

#pragma endregion

#pragma region query string part 3 (where)

	queryString += " WHERE ";
	for (auto tuple = filter.begin(); tuple != filter.end(); ++tuple)
	{
		queryString += tuple.key() + "=" + "'" + tuple.value() + "'";
		if (tuple != filter.end() - 1) { queryString += " AND "; }
	}

#pragma endregion

	// execute and return
	return executeMySQL(queryString);
}

QString& MySqlWorker::DatabaseWorker::getLastError()
{
	return lastError;
}

void MySqlWorker::DatabaseWorker::setLastError(QString additionalError)
{
	lastError = QSqlDatabase::database(this->databaseName).lastError().text();
	if (additionalError != "") { lastError += "\n" + additionalError; }
}

bool MySqlWorker::DatabaseWorker::executeMySQL(QString &commandString)
{
	QSqlDatabase db = QSqlDatabase::database(this->databaseName);
	bool sucess = db.open();

	QSqlQuery query(db);

	sucess &= query.exec(commandString);
	if(!sucess) { setLastError(query.lastError().text()); }
	db.close();
	return sucess;
}
